# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

"""
This module defines the behaviors of ``TGraphErrors``.
"""


class TGraphErrors:
    """
    Behaviors for TGraphErrors: get values and errors as NumPy arrays.
    """

    # TO DO!
